import java.util.*;

// Represents a self-balancing AVLTree in which the guaranteed height is max log(n)
public class StringAVLTree implements Searchable  {
    private AVLNode overallRoot;

    // Post: Constructs an empty AVL tree
    public StringAVLTree() {
        this.overallRoot = null;
    }

    // Post: Copies all the provided values to this AVL tree, clearing out any previously
    //       stored values
    public void copy(List<String> values) {
        this.overallRoot = null;
        for (String x : values) {
            this.add(x);
        }
    }

    // Post: Returns whether or not the provided value is present within this AVL tree
    public boolean contains(String value) {
        return contains(value, overallRoot);
    }

    // Post: Returns whether or not the provided value is present within the AVL tree rooted
    //       at the provided root
    private boolean contains(String value, AVLNode root) {
        if (root == null) {
            return false;
        } else if (root.data.equals(value)) {
            return true;
        } else if (value.compareTo(root.data) < 0) {
            return contains(value, root.left);
        } else {
            return contains(value, root.right);
        }
    }

    // Post: Adds the provided value to this AVL tree
    public void add(String value) {
        this.overallRoot = insert(value, overallRoot);
    }

    // Post: Adds the provided value into the AVL tree rooted at curr
    private AVLNode insert(String value, AVLNode curr) {
        if (curr == null) {
            return new AVLNode(value);
        } else {
            if (value.compareTo(curr.data) < 0) {
                curr.left = insert(value, curr.left);
            } else if (value.compareTo(curr.data) > 0) {
                curr.right = insert(value, curr.right);
            }

            int left = getHeight(curr.left);
            int right = getHeight(curr.right);
            boolean isUnbalanced = Math.abs(left - right) >= 2;
            if (isUnbalanced) {
                boolean firstLeft = left > right;
                AVLNode problemChild = firstLeft ? curr.left : curr.right;

                int doubleLeft = getHeight(problemChild.left);
                int doubleRight = getHeight(problemChild.right);
                boolean secondLeft = doubleLeft > doubleRight;

                curr = (firstLeft && secondLeft || !firstLeft && !secondLeft) ?
                       singleRotate(curr, !firstLeft) :
                       doubleRotate(curr, !firstLeft);
            } else {
                curr.height = Math.max(left, right) + 1;
            }
        }
        return curr;
    }

    // Post: Performs a single rotation to balance the AVLTree
    private AVLNode singleRotate(AVLNode root, boolean left) {
        AVLNode temp = null;
        if (!left) {     //Case 1 (rotate right)
            temp = root.left;
            root.left = temp.right;
            temp.right = root;
        } else {        //Case 4 (rotate left)
            temp = root.right;
            root.right = temp.left;
            temp.left = root;
        }
        root.height = Math.max(getHeight(root.left), getHeight(root.right)) + 1;
        temp.height = Math.max(getHeight(temp.left), getHeight(temp.right)) + 1;
        return temp;
    }

    // Post: Performs a double rotation to balance the AVLTree
    private AVLNode doubleRotate(AVLNode root, boolean right) {
        if (!right) {     //Case 3 (rotate right)
            root.left = singleRotate(root.left, true);
        } else {        //Case 2 (rotate left)
            root.right = singleRotate(root.right, false);
        }
        root = singleRotate(root, right);
        return root;
    }

    // Post: Returns the height of the given node, -1 if null
    private int getHeight(AVLNode node) {
        return node == null ? -1 : node.height;
    }

    // Class that represents a single node in the AVL tree
    private static class AVLNode {
        public int height;  // height of this node in the tree
        public String data;    // data stored at this node
        public AVLNode left;    // reference to left subtree
        public AVLNode right;   // reference to right subtree

        // Post: Constructs a new AVL node with the given data
        public AVLNode(String data) {
            this(data, null, null);
        }

        // Post: Constructs a new AVL node with the given data and links.
        public AVLNode(String data, AVLNode left, AVLNode right) {
            this.data = data;
            this.left = left;
            this.right = right;
            this.height = 0;
        }
    }
}
